#!/usr/bin/env python3
from __future__ import annotations

import argparse
import logging
import os
import re
from pathlib import Path
from typing import cast

import parsy as ps
from colorama import Back, Fore, Style
from parsy import Parser, generate, peek, regex, seq, string

LEVEL_UNKNOWN = 255
NEWLINE = string("\n")
SPACE = string(" ")


class Config(dict[str, str]):
    class LevelAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None) -> None:
            _ = (parser, option_string)
            values = cast(str, values)
            setattr(namespace, self.dest, get_debug_level_const(values.upper())[0])

    def __init__(self) -> None:
        self.dict = {}
        if os.path.exists(".env"):
            from dotenv import dotenv_values

            self.dict = {
                k: v for k, v in dotenv_values(".env").items() if v is not None
            }
        parser = argparse.ArgumentParser()
        if not ("LOGFILE" in self.dict or "LOGFILE" in os.environ):
            parser.add_argument(
                "LOGFILE",
                type=Path,
            )
        parser.add_argument("-q", "--query", dest="QUERY")
        parser.add_argument(
            "-l",
            "--level",
            dest="LOGLEVEL",
            type=str,
            choices=["debug", "info", "warn", "err", "crit", "unk"],
            default=logging.INFO,
            action=self.LevelAction,
        )

        self.dict.update(parser.parse_args().__dict__)

    def __getitem__(self, key: str, /) -> str:
        try:
            return self.dict[key]
        except KeyError:
            return os.environ[key]

    def __setitem__(self, _key: str, _value: str, /) -> None:
        msg = "Config is read-only"
        raise NotImplementedError(msg)

    def __delitem__(self, _: str, /) -> None:
        msg = "Config is read-only"
        raise NotImplementedError(msg)

    def __iter__(self):
        return iter(self.dict)

    def __len__(self) -> int:
        return len(self.dict)

    def __repr__(self) -> str:
        return self.dict.__repr__()


"""
Formatting functions
"""


def colorize_status(status: str) -> str:
    if "PASS" in status:
        return Fore.GREEN + status + Style.RESET_ALL
    elif "FAIL" in status:
        return Fore.RED + status + Style.RESET_ALL
    elif "ERROR" in status:
        return Fore.BLACK + Back.RED + status + Style.RESET_ALL
    return status


def get_debug_level_const(level: str) -> tuple[int, str]:
    if "DEBUG" in level:
        return (logging.DEBUG, level)
    elif "INFO" in level:
        return (logging.INFO, level)
    elif "WARN" in level:
        return (logging.WARNING, level)
    elif "ERR" in level:
        return (logging.ERROR, level)
    elif "CRIT" in level:
        return (logging.CRITICAL, level)
    else:
        return (LEVEL_UNKNOWN, level)


def colorize_by_level(level: str, text: str) -> str:
    if "DEBUG" in level:
        return Style.DIM + text + Style.RESET_ALL
    elif "INFO" in level:
        return Style.BRIGHT + text + Style.RESET_ALL
    elif "WARN" in level:
        return Fore.YELLOW + text + Style.RESET_ALL
    elif "ERR" in level:
        return Fore.RED + text + Style.RESET_ALL
    elif "CRIT" in level:
        return Style.BRIGHT + Back.RED + Fore.BLACK + text + Style.RESET_ALL
    else:
        return Fore.BLUE + text + Style.RESET_ALL
    # return level


"""
Output functions
"""


class TestPrinterBuilder:
    loglevel: int = logging.NOTSET
    query: str | re.Pattern[str] | None = None

    def __call__(
        self, ts: str, level: tuple[int, str], src: str, msg: str
    ) -> str | None:
        level_name = level[1]
        if (level[0] >= self.loglevel) or (
            self.query is not None and re.search(self.query, msg)
        ):
            res = f"{ts}\t{Fore.BLACK + Back.WHITE + level[1] + Style.RESET_ALL}\t{colorize_by_level(level_name,f'{src}:    {msg}')}"

            print(res)
            return res

    def with_loglevel(self, loglevel: int) -> TestPrinterBuilder:
        self.loglevel = loglevel
        return self

    def with_query(self, query: str | re.Pattern) -> TestPrinterBuilder:
        self.query = query
        return self


"""
Parser defs
"""


def block_parser(text: str) -> Parser:
    return regex(rf"^.*\n^.*{text}.*$\n.*", re.MULTILINE)


timestamp: Parser = (
    regex(
        r"^\d{4}-\d{2}-\d{2}.\d{2}:\d{2}:\d{2}(\+\d{2}:\d{2})?(,\d{3})?", re.MULTILINE
    )
    << string(":").optional()
)
debug_level: Parser = regex(r"[A-Za-z0-9_%-]+") << string(":").optional()


start_block = block_parser("EASYPY JOB START")
end_block = block_parser("EASYPY JOB END")
report_block = block_parser("Easypy Report")
summary_block = block_parser("Task Result Summary")
details_block = block_parser("Task Result Details")


message = ps.any_char.until(ps.peek(timestamp) | ps.eof).concat()


@generate
def test_output_line():
    ts = yield timestamp << SPACE
    level = yield debug_level << SPACE
    src = yield (string("[") >> regex(r"[A-Za-z_0-9]+") << string("] ")).optional()
    msg = yield message.map(lambda x: x.rstrip())

    return print_test_line(ts, get_debug_level_const(level), src, msg)

    # return {"ts": ts, "level": level, "src": src, "msg": msg}


@generate
def test_details():
    test_name = yield (
        (
            NEWLINE.optional()
            << (SPACE * 4)
            << (string("|--") | string("`--"))
            << ps.whitespace
        )
        >> ps.any_char.until(ps.whitespace).concat()
        << ps.whitespace
    )
    test_results = yield (ps.letter.many().concat())
    print(f" |  {colorize_status(test_results)}\t{test_name}")
    return {"test_results": test_results, "test_name": test_name}


@generate
def subtask_details():
    subtask_name = (
        yield (string("`--") << ps.whitespace)
        >> ps.any_char.until(ps.whitespace).concat()
        << ps.whitespace
    )
    subtask_result = yield (ps.letter.many().concat())
    print(f" > {colorize_status(subtask_result)}\t{subtask_name}")
    tests = yield test_details.many()
    return {
        "subtask_result": subtask_result,
        "subtask_name": subtask_name,
        "tests": tests,
    }


@generate
def task_details():
    task_number = (
        yield ps.any_char.until(string(":")).concat() << string(":") << ps.whitespace
    )
    task_name = yield (ps.any_char.until(ps.whitespace).concat() << ps.whitespace)
    print(f"{task_number}\t {Style.BRIGHT + task_name + Style.RESET_ALL}")

    subtasks = yield (subtask_details.many())
    return {"task_number": task_number, "task_name": task_name, "subtasks": subtasks}


@generate
# NOTE: Uncomment the print statements if you actually care what's in these sections
def sections():
    header = yield (
        ps.any_char.until(peek(timestamp)).concat()  # TODO: output this
        + test_output_line.until(start_block).concat()
        << start_block
        << NEWLINE
    )

    test_output = yield test_output_line.until(end_block) << end_block << NEWLINE
    post_test = yield (ps.any_char.until(report_block).concat() << report_block)
    # print(post_test)
    report = yield (ps.any_char.until(summary_block).concat() << summary_block)
    # print(report)
    summary = yield (ps.any_char.until(details_block).concat() << details_block)
    # print(summary)
    details = yield task_details.many()
    etc = yield (ps.any_char.until(ps.eof).concat())
    return {
        "header": header,
        "test_output": test_output,
        "post_test": post_test,
        "report": report,
        "summary": summary,
        "details": details,
        "etc": etc,
    }


# TODO: make this work with streams
if __name__ == "__main__":
    config = Config()
    with Path(config["LOGFILE"]).open() as f:
        log = f.read()
    global output_test_line
    print_test_line = TestPrinterBuilder()
    if loglevel := config.get("LOGLEVEL"):
        print_test_line = print_test_line.with_loglevel(cast(int, loglevel))
    if query := config.get("QUERY"):
        print_test_line = print_test_line.with_query(query)
    from icecream import ic

    parsed_sections = sections.parse(log)
